from typing import Any, Dict, List

from mat3ra.ade.context.context_provider import ContextProvider
from mat3ra.esse.models.context_providers_directory.points_grid_data_provider import (
    GridMetricType,
    PointsGridDataProviderSchema,
)
from pydantic import Field


# TODO: GlobalSetting for default KPPRA value
class PointsGridDataProvider(PointsGridDataProviderSchema, ContextProvider):
    """
    Context provider for k-point/q-point grid configuration.

    Handles grid dimensions and shifts for reciprocal space sampling.
    """

    name: str = Field(default="kgrid")
    divisor: int = Field(default=1)
    dimensions: List[int] = Field(default_factory=lambda: [1, 1, 1])
    shifts: List[float] = Field(default_factory=lambda: [0.0, 0.0, 0.0])
    grid_metric_type: str = Field(default=GridMetricType.KPPRA)

    @property
    def is_edited_key(self) -> str:
        return "isKgridEdited"

    @property
    def default_data(self) -> Dict[str, Any]:
        return {
            "dimensions": self.dimensions,
            "shifts": self.shifts,
            "gridMetricType": self.grid_metric_type,
            "divisor": self.divisor,
        }

    # TODO: add a test to verify context and templates are the same as from JS implementation
    def get_default_grid_metric_value(self, metric: str) -> float:
        raise NotImplementedError

    def calculate_dimensions(
            self, grid_metric_type: str, grid_metric_value: float, units: str = "angstrom"
    ) -> List[int]:
        raise NotImplementedError

    def calculate_grid_metric(self, grid_metric_type: str, dimensions: List[int], units: str = "angstrom") -> float:
        raise NotImplementedError

    def transform_data(self, data: Dict[str, Any]) -> Dict[str, Any]:
        raise NotImplementedError



from typing import Optional, get_args

from mat3ra.esse.models.workflow.unit.subworkflow import SubworkflowUnitSchema
from pydantic import Field

from .unit import Unit


class SubworkflowUnit(Unit, SubworkflowUnitSchema):
    type: Optional[str] = Field(default=get_args(SubworkflowUnitSchema.model_fields["type"].annotation)[0])



from mat3ra.ade.context.context_provider import ContextProvider


class ExecutableContextProvider(ContextProvider):
    """
    Context provider for executable settings.
    """

    domain: str = "executable"



from typing import List, Optional, TypeVar

from mat3ra.utils import find_by_key_or_regex

from ..units import Unit

T = TypeVar("T")


class FlowchartUnitsManager:
    """
    Mixin class providing common unit operations for flowchart units.

    This mixin expects the class to have a `units: List[Unit]` attribute.
    It provides common methods for managing units in both Workflow and Subworkflow classes.
    """

    units: List[Unit]

    def set_units(self, units: List[Unit]) -> None:
        self.units = units

    def get_unit(self, flowchart_id: str) -> Optional[Unit]:
        for unit in self.units:
            if unit.flowchartId == flowchart_id:
                return unit
        return None

    def find_unit_by_id(self, id: str) -> Optional[Unit]:
        for unit in self.units:
            if getattr(unit, "id", None) == id:
                return unit
        return None

    def find_unit_with_tag(self, tag: str) -> Optional[Unit]:
        for unit in self.units:
            if hasattr(unit, "tags") and unit.tags is not None and tag in unit.tags:
                return unit
        return None

    def get_unit_by_name(
        self,
        name: Optional[str] = None,
        name_regex: Optional[str] = None,
    ) -> Optional[Unit]:
        return find_by_key_or_regex(self.units, key="name", value=name, value_regex=name_regex)

    @staticmethod
    def _add_to_list(items: List[T], item: T, head: bool = False, index: int = -1) -> None:
        """
        Add an item to a list at a specified position.

        Args:
            items: The list to add to
            item: The item to add
            head: If True, insert at the beginning (index 0)
            index: If >= 0, insert at this specific index
                   If < 0, append to the end
        """
        if head:
            items.insert(0, item)
        elif index >= 0:
            items.insert(index, item)
        else:
            items.append(item)

    def set_units_head(self, units: List[Unit]) -> List[Unit]:
        """
        Set the head flag on the first unit and unset it on all others.

        Args:
            units: List of units to process

        Returns:
            The modified units list
        """
        if len(units) > 0:
            units[0].head = True
            for unit in units[1:]:
                unit.head = False
        return units

    def set_next_links(self, units: List[Unit]) -> List[Unit]:
        """
        Re-establishes the linked next => flowchartId logic in an array of units.

        Args:
            units: List of units to process

        Returns:
            The modified units list
        """
        flowchart_ids = [unit.flowchartId for unit in units]

        for i in range(len(units) - 1):
            unit_next = getattr(units[i], "next", None)

            if unit_next is None:
                units[i].next = units[i + 1].flowchartId
                if i > 0:
                    units[i - 1].next = units[i].flowchartId
            elif unit_next not in flowchart_ids:
                units[i].next = units[i + 1].flowchartId

        return units

    def _clear_link_to_unit(self, flowchart_id: str) -> None:
        """
        Clear the 'next' link from any unit that points to the given flowchart_id.

        This is used to mend broken links when removing a unit.

        Args:
            flowchart_id: The flowchart_id to clear links to
        """
        for unit in self.units:
            if getattr(unit, "next", None) == flowchart_id:
                unit.next = None
                break

    def add_unit(self, unit: Unit, head: bool = False, index: int = -1) -> None:
        """
        Add a unit to the units list.

        Args:
            unit: Unit to add
            head: If True, add at the beginning
            index: If >= 0, insert at this index
        """
        if len(self.units) == 0:
            unit.head = True
            self.set_units([unit])
        else:
            self._add_to_list(self.units, unit, head, index)
            self.set_units(self.set_next_links(self.set_units_head(self.units)))

    def remove_unit(self, flowchart_id: str) -> None:
        """
        Remove a unit by its flowchartId.

        Args:
            flowchart_id: The flowchartId of the unit to remove
        """

        if len(self.units) < 2:
            return

        unit_to_remove = None
        for unit in self.units:
            if unit.flowchartId == flowchart_id:
                unit_to_remove = unit
                break

        if not unit_to_remove:
            return

        self._clear_link_to_unit(unit_to_remove.flowchartId)

        remaining_units = [unit for unit in self.units if unit.flowchartId != flowchart_id]
        units_with_head = self.set_units_head(remaining_units)
        self.units = self.set_next_links(units_with_head)

    def replace_unit(self, index: int, unit: Unit) -> None:
        """
        Replace a unit at a specific index.

        Args:
            index: Index of the unit to replace
            unit: New unit to place at that index
        """

        if 0 <= index < len(self.units):
            self.units[index] = unit
            self.set_units(self.set_next_links(self.set_units_head(self.units)))

    def set_unit(
        self,
        new_unit: Unit,
        unit: Optional[Unit] = None,
        unit_flowchart_id: Optional[str] = None,
    ) -> bool:
        """
        Replace a unit by finding it either by instance or flowchart_id.
        Unit can replace itself if it was modified outside the class.

        Args:
            new_unit: The new unit to set
            unit: The existing unit instance to replace (optional)
            unit_flowchart_id: The flowchart_id of the unit to replace (optional)

        If neither unit nor unit_flowchart_id is provided, the function will use
        new_unit.flowchartId to find the existing unit to replace.

        Returns:
            True if successful, False otherwise
        """
        if unit is not None:
            target_unit = unit
        elif unit_flowchart_id is not None:
            target_unit = self.get_unit(unit_flowchart_id)
        else:
            target_unit = self.get_unit(new_unit.flowchartId)

        if target_unit is None:
            return False

        try:
            unit_index = self.units.index(target_unit)
            self.replace_unit(unit_index, new_unit)
            return True
        except ValueError:
            return False



from typing import Any, Dict, List

from mat3ra.code.entity import InMemoryEntitySnakeCase
from mat3ra.code.mixins import HashedEntityMixin
from mat3ra.esse.models.workflow.unit.base import WorkflowBaseUnitSchema
from mat3ra.utils.uuid import get_uuid
from pydantic import Field


class Unit(WorkflowBaseUnitSchema, HashedEntityMixin, InMemoryEntitySnakeCase):
    """
    Unit class representing a unit of computational work in a workflow.

    Attributes:
        type: Type of the unit (e.g., execution, assignment, condition)
        name: Name of the unit
        flowchartId: Unique identifier for the unit in the flowchart
        head: Whether this unit is the head of the workflow
        next: Flowchart ID of the next unit
        tags: List of tags for the unit
        context: Context data dictionary for the unit
    """
    id: str = Field(default_factory=get_uuid, alias="_id")
    flowchartId: str = Field(default_factory=get_uuid)
    # TODO: use RuntimeItemNameObjectSchema when available
    preProcessors: List[Any] = Field(default_factory=list)
    postProcessors: List[Any] = Field(default_factory=list)
    monitors: List[Any] = Field(default_factory=list)
    results: List[Any] = Field(default_factory=list)
    context: Dict[str, Any] = Field(default_factory=dict)


    def get_hash_object(self) -> Dict[str, Any]:
        return {
            "results": self.results or [],
            "preProcessors": self.preProcessors or [],
            "postProcessors": self.postProcessors or [],
            "type": self.type,
        }

    def is_in_status(self, status: str) -> bool:
        return self.status == status

    def add_context(self, new_context: Dict[str, Any]):
        self.context.update(new_context)

    def set_context(self, new_context: Dict[str, Any]):
        self.context = new_context

    def get_context(self, key: str, default: Any = None) -> Any:
        return self.context.get(key, default)

    def remove_context(self, key: str):
        self.context.pop(key, None)

    def clear_context(self):
        self.context = {}



import pytest
from mat3ra.standata.applications import ApplicationStandata
from mat3ra.standata.workflows import WorkflowStandata
from mat3ra.wode import Unit

WORKFLOW_STANDATA = WorkflowStandata()
APPLICATION_STANDATA = ApplicationStandata()

DEFAULT_WF_NAME = WORKFLOW_STANDATA.get_default()["name"]
APPLICATION_ESPRESSO = APPLICATION_STANDATA.get_by_name_first_match("espresso")["name"]

UNIT_FLOWCHART_ID = "abc-123-def"
UNIT_NEXT_ID = "next-456"

NEW_CONTEXT_RELAX = {"kgrid": {"density": 0.5}, "convergence": {"threshold": 1e-6}}

UNIT_CONFIG_EXECUTION = {
    "type": "execution",
    "name": "pw_scf",
    "flowchartId": UNIT_FLOWCHART_ID,
    "head": True,
}

UNIT_CONFIG_ASSIGNMENT = {
    "type": "assignment",
    "name": "kgrid",
    "flowchartId": "kgrid-flowchart-id",
    "head": False,
}


@pytest.mark.parametrize("config", [UNIT_CONFIG_EXECUTION, UNIT_CONFIG_ASSIGNMENT])
def test_creation(config):
    unit = Unit(**config)
    assert unit.type == config["type"]
    assert unit.name == config["name"]


def test_snake_case_properties():
    unit = Unit(**UNIT_CONFIG_EXECUTION)
    assert unit.flowchart_id == UNIT_FLOWCHART_ID


@pytest.mark.parametrize("head_value", [True, False])
def test_head_property(head_value):
    config = {**UNIT_CONFIG_EXECUTION, "head": head_value}
    unit = Unit(**config)
    assert unit.head == head_value


def test_next_property():
    config = {**UNIT_CONFIG_EXECUTION, "next": UNIT_NEXT_ID}
    unit = Unit(**config)
    assert unit.next == UNIT_NEXT_ID


def test_add_context():
    unit = Unit(**{**UNIT_CONFIG_EXECUTION, "name": "relaxation step"})

    assert unit is not None
    assert "relax" in unit.name.lower()

    unit.add_context(NEW_CONTEXT_RELAX)

    assert "kgrid" in unit.context
    assert "convergence" in unit.context
    assert unit.context["kgrid"] == NEW_CONTEXT_RELAX["kgrid"]
    assert unit.context["convergence"] == NEW_CONTEXT_RELAX["convergence"]



# TODO: We need periodic_table.js equivalent in Python
# TODO: We need all mixins equivalent in Python

from typing import Any, List

from mat3ra.esse.models.context_providers_directory.by_application.qe_pwx_context_provider import (
    QEPwxContextProviderSchema,
)

from ..executable_context_provider import ExecutableContextProvider


class QEPWXContextProvider(QEPwxContextProviderSchema, ExecutableContextProvider):
    """
    Context provider for Quantum ESPRESSO pw.x settings.
    """

    # self.init_materials_context_mixin()
    # self.init_method_data_context_mixin()
    # self.init_workflow_context_mixin()
    # self.init_job_context_mixin()
    # self.init_material_context_mixin()
    _material: Any = None
    _materials: List[Any] = []



from typing import Any, Dict, List, Literal

from mat3ra.ade import Application, Executable, Flavor
from mat3ra.esse.models.workflow.unit.execution import ExecutionUnitSchemaBase
from mat3ra.utils import (
    calculate_hash_from_object,
    remove_comments_from_source_code,
    remove_empty_lines_from_string,
    remove_timestampable_keys,
)
from pydantic import Field

from .unit import Unit


class ExecutionUnit(Unit, ExecutionUnitSchemaBase):
    type: Literal["execution"] = "execution"
    executable: Executable = None
    flavor: Flavor = None
    application: Application = None
    input: List = Field(default_factory=list)

    def get_hash_object(self) -> Dict[str, Any]:
        app = self.application.to_dict() if self.application else {}
        exe = self.executable.to_dict() if self.executable else {}
        flv = self.flavor.to_dict() if self.flavor else {}
        input_items = self.input if isinstance(self.input, list) else []
        input_hash = calculate_hash_from_object(
            [
                remove_empty_lines_from_string(remove_comments_from_source_code(i.get("content", "")))
                for i in input_items
                if isinstance(i, dict)
            ]
        )
        return {
            **super().get_hash_object(),
            "application": remove_timestampable_keys(app),
            "executable": remove_timestampable_keys(exe),
            "flavor": remove_timestampable_keys(flv),
            "input": input_hash,
        }



from typing import Any, Dict, List, Optional

from mat3ra.code.entity import InMemoryEntitySnakeCase
from mat3ra.code.mixins import HashedEntityMixin
from mat3ra.esse.models.workflow import WorkflowSchema
from mat3ra.standata.subworkflows import SubworkflowStandata
from mat3ra.utils.uuid import get_uuid
from pydantic import Field

from ..mixins import FlowchartUnitsManager
from ..subworkflows import Subworkflow
from ..units import Unit


class Workflow(WorkflowSchema, HashedEntityMixin, InMemoryEntitySnakeCase, FlowchartUnitsManager):
    """
    Workflow class representing a complete workflow configuration.

    Attributes:
        name: Name of the workflow
        subworkflows: List of subworkflows in the workflow
        units: List of units linking the subworkflows
        properties: List of properties extracted by the workflow
    """

    id: str = Field(default_factory=get_uuid, alias="_id")
    subworkflows: List[Subworkflow] = Field(default_factory=list)
    units: List[Unit] = Field(default_factory=list)
    isMultiMaterial: bool = Field(default=False)

    @property
    def application(self):
        if not self.subworkflows or len(self.subworkflows) == 0:
            return None

        first_subworkflow = self.subworkflows[0]
        return first_subworkflow.application if first_subworkflow.application else None

    @property
    def relaxation_subworkflow(self) -> Optional[Subworkflow]:
        application_name = self.application.name if self.application else None
        subworkflow_standata = SubworkflowStandata()
        relaxation_data = subworkflow_standata.get_relaxation_by_application(application_name)
        return Subworkflow(**relaxation_data) if relaxation_data else None

    @property
    def has_relaxation(self) -> bool:
        return self._find_relaxation_subworkflow() is not None

    def add_subworkflow(self, subworkflow: Subworkflow, head: bool = False, index: int = -1):
        self._add_to_list(self.subworkflows, subworkflow, head, index)
        unit = subworkflow.get_as_unit()
        self.add_unit(unit, head, index)

    def remove_subworkflow_by_id(self, id: str):
        self.subworkflows = [sw for sw in self.subworkflows if sw.id != id]

    def find_subworkflow_by_id(self, id: str) -> Optional[Subworkflow]:
        return next((sw for sw in self.subworkflows if sw.id == id), None)

    def set_context_to_unit(
        self,
        unit_name: Optional[str] = None,
        unit_name_regex: Optional[str] = None,
        new_context: Optional[Dict[str, Any]] = None,
    ):
        target_unit = self.get_unit_by_name(name=unit_name, name_regex=unit_name_regex)
        target_unit.context = new_context

    def _find_relaxation_subworkflow(self) -> Optional[Subworkflow]:
        target_name = self.relaxation_subworkflow.name

        return next(
            (swf for swf in self.subworkflows if swf.name == target_name),
            None,
        )

    def add_relaxation(self) -> None:
        if self.has_relaxation:
            return

        relaxation_definition = self.relaxation_subworkflow
        if relaxation_definition is not None:
            self.add_subworkflow(relaxation_definition, head=True)

    def get_hash_object(self) -> Dict[str, Any]:
        nested_workflows = getattr(self, "workflows", []) or []
        return {
            "units": ",".join(u.calculate_hash() for u in self.units),
            "subworkflows": ",".join(sw.calculate_hash() for sw in self.subworkflows),
            "workflows": ",".join(w.calculate_hash() for w in nested_workflows),
        }

    def remove_relaxation(self) -> None:
        existing = self._find_relaxation_subworkflow()
        if existing is not None:
            self.remove_subworkflow_by_id(existing.id)



from typing import Any, List, Optional, Union

from mat3ra.ade.application import Application
from mat3ra.code.entity import InMemoryEntitySnakeCase
from mat3ra.code.mixins import HashedEntityMixin
from mat3ra.esse.models.workflow.subworkflow import Subworkflow as SubworkflowSchema
from mat3ra.mode.method import Method
from mat3ra.mode.model import Model
from mat3ra.utils.uuid import get_uuid
from pydantic import Field, field_validator

from ..mixins import FlowchartUnitsManager
from ..units import ExecutionUnit, SubworkflowUnit, Unit
from ..units.builders import build_execution_unit_config


class Subworkflow(SubworkflowSchema, HashedEntityMixin, InMemoryEntitySnakeCase, FlowchartUnitsManager):
    """
    Subworkflow class representing a logical collection of workflow units.

    Attributes:
        name: Name of the subworkflow
        application: Application configuration
        model: Model configuration
        units: List of units in the subworkflow
        properties: List of properties extracted by the subworkflow
    """

    id: str = Field(default_factory=get_uuid, alias="_id")
    application: Application = Field(
        default_factory=lambda: Application(name="", version="", build="", shortName="", summary="")
    )
    model: Model = Field(default_factory=lambda: Model(type="", subtype="", method=Method(type="", subtype="")))
    units: List[Union[Unit, ExecutionUnit, SubworkflowUnit]] = Field(default_factory=list)

    @field_validator("units", mode="before")
    @classmethod
    def _instantiate_units(cls, value: Any) -> Any:
        items = value if isinstance(value, list) else []
        instantiated: List[Unit] = []
        for item in items:
            if isinstance(item, Unit):
                instantiated.append(item)
                continue
            if not isinstance(item, dict):
                continue
            unit_type = item.get("type")
            if unit_type == "execution":
                instantiated.append(ExecutionUnit(**build_execution_unit_config(item)))
            elif unit_type == "subworkflow":
                instantiated.append(SubworkflowUnit(**item))
            else:
                instantiated.append(Unit(**item))
        return instantiated

    @classmethod
    def from_arguments(
        cls,
        application: Application,
        model: Model,
        method: Method,
        name: str,
        units: Optional[List] = None,
        config: Optional[dict] = None,
    ) -> "Subworkflow":
        if units is None:
            units = []
        if config is None:
            config = {}

        model.method = method
        return cls(name=name, application=application, model=model, units=units, **config)

    @property
    def method_data(self):
        return self.model.method.data

    def get_hash_object(self) -> dict:
        app_dict = self.application.to_dict() if self.application else {}
        model_dict = self.model.to_dict() if self.model else {}
        return {
            "application": Application.create(app_dict).calculate_hash(),
            "model": Model.create(model_dict).calculate_hash(),
            "units": ",".join(u.calculate_hash() for u in self.units),
        }

    def get_as_unit(self) -> Unit:
        return Unit(type="subworkflow", id=self.id, name=self.name)



from .mixins import FlowchartUnitsManager
from .subworkflows import Subworkflow
from .units import ExecutionUnit, SubworkflowUnit, Unit
from .workflows import Workflow

__all__ = [
    "Unit",
    "ExecutionUnit",
    "SubworkflowUnit",
    "Subworkflow",
    "Workflow",
    "FlowchartUnitsManager",
]



import pytest
from mat3ra.ade.application import Application
from mat3ra.mode.method import Method
from mat3ra.mode.model import Model
from mat3ra.standata.applications import ApplicationStandata
from mat3ra.wode import Subworkflow, Unit

SUBWORKFLOW_NAME = "Total Energy"
SUBWORKFLOW_APPLICATION = Application(**ApplicationStandata.get_by_name_first_match("espresso"))
SUBWORKFLOW_METHOD = Method(type="pseudopotential", subtype="us")
SUBWORKFLOW_MODEL = Model(type="dft", subtype="gga", method=SUBWORKFLOW_METHOD)
SUBWORKFLOW_PROPERTIES = ["total_energy", "pressure"]

UNIT_CONFIG = {
    "type": "execution",
    "name": "pw_scf",
    "flowchartId": "unit-flowchart-id",
    "head": True,
}


def test_creation():
    sw = Subworkflow(name=SUBWORKFLOW_NAME)
    assert sw.name == SUBWORKFLOW_NAME


@pytest.mark.parametrize("app_name", ["espresso", "vasp"])
def test_application(app_name):
    app_data = ApplicationStandata.get_by_name_first_match(app_name)
    application = Application(**app_data)
    sw = Subworkflow(name=SUBWORKFLOW_NAME, application=application)
    assert sw.application.name == app_name
    assert sw.application.version == app_data["version"]


@pytest.mark.parametrize(
    "model_type,model_subtype",
    [
        ("dft", "gga"),
        ("dft", "lda"),
    ],
)
def test_model(model_type, model_subtype):
    method = Method(type="pseudopotential", subtype="us")
    model = Model(type=model_type, subtype=model_subtype, method=method)
    sw = Subworkflow(name=SUBWORKFLOW_NAME, model=model)
    assert sw.model.type == model_type
    assert sw.model.subtype == model_subtype


def test_with_units():
    unit = Unit(**UNIT_CONFIG)
    sw = Subworkflow(name=SUBWORKFLOW_NAME, units=[unit])
    assert len(sw.units) == 1
    assert sw.units[0].name == UNIT_CONFIG["name"]


def test_id_generation():
    sw1 = Subworkflow(name=SUBWORKFLOW_NAME)
    sw2 = Subworkflow(name=SUBWORKFLOW_NAME)
    assert sw1.id != sw2.id

def test_get_as_unit():
    sw = Subworkflow(name=SUBWORKFLOW_NAME)
    unit = sw.get_as_unit()
    assert unit.type == "subworkflow"
    assert unit.id == sw.id
    assert unit.to_dict().get("_id") == sw.id
    assert unit.name == sw.name



from __future__ import annotations

from typing import Any, Dict

from mat3ra.standata.applications import ApplicationStandata


def build_execution_unit_config(config: Dict[str, Any]) -> Dict[str, Any]:
    app = config.get("application") if isinstance(config.get("application"), dict) else {}
    exe = config.get("executable") if isinstance(config.get("executable"), dict) else {}

    app_name = app.get("name")
    exe_name = exe.get("name")
    if not app_name or not exe_name or exe.get("results"):
        return config

    exe_full = ApplicationStandata.get_executable_flavor_map_by_application_name(app_name).get(exe_name) or {}
    results = exe_full.get("results") if isinstance(exe_full, dict) else None
    if not results:
        return config

    updated = dict(config)
    updated["executable"] = {**exe, "results": results}
    return updated
